'use strict';

var path = require('path');
var fs = require('fs');
var crypto = require('crypto');
var Util = require('./Util.js');

var classCallCheck = function (instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
};

var createClass = function () {
  function defineProperties(target, props) {
    for (var i = 0; i < props.length; i++) {
      var descriptor = props[i];
      descriptor.enumerable = descriptor.enumerable || false;
      descriptor.configurable = true;
      if ("value" in descriptor) descriptor.writable = true;
      Object.defineProperty(target, descriptor.key, descriptor);
    }
  }

  return function (Constructor, protoProps, staticProps) {
    if (protoProps) defineProperties(Constructor.prototype, protoProps);
    if (staticProps) defineProperties(Constructor, staticProps);
    return Constructor;
  };
}();

/**
 * static file server module
*/

var StaticFileServer = function () {

    /**
     *@param {string} rootDir - the project root directory
     *@param {Array} publicPaths - array of public paths to serve static files from
     *@param {Object} mimeTypes - object of file mime types
     *@param {Array} defaultDocuments - array of folder default documents
     *@param {string} cacheControl - cache control header for static files
    */
    function StaticFileServer(rootDir, publicPaths, mimeTypes, defaultDocuments, cacheControl) {
        classCallCheck(this, StaticFileServer);

        this.rootDir = rootDir;
        this.publicPaths = publicPaths.map(function (publicPath) {
            return path.join(rootDir, publicPath, '/');
        });
        this.mimeTypes = mimeTypes;
        this.defaultDocuments = defaultDocuments;
        this.cacheControl = cacheControl || 'no-cache, max-age=86400';
    }

    /**
     * return instance identity
    */


    createClass(StaticFileServer, [{
        key: 'getDefaultHeaders',


        /**
         * returns default response headers
         *@param {string} filePath - the file path
         *@returns {Object}
        */
        value: function getDefaultHeaders(filePath) {
            var stat = fs.statSync(filePath);

            return {
                'Content-Type': this.mimeTypes[path.parse(filePath).ext.substring(1)] || 'application/octet-stream',
                'Last-Modified': stat.mtime.toString(),
                'Content-Length': stat.size,
                'ETag': this.getFileTag(stat.mtime),
                'Cache-Control': this.cacheControl
            };
        }

        /**
         * ends the response.
         *@param {RServerResponse} response - the response object
         *@param {number} status - response status code
         *@param {Object} headers - the response headers to write
         *@param {string|Buffer} [data] - response data to send
         *@param {function} [callback] - a callback method
         *@returns {boolean}
        */

    }, {
        key: 'endResponse',
        value: function endResponse(response, status, headers, data, callback) {
            callback = Util.isCallable(callback) ? callback : function () {};
            response.writeHead(status, headers || {});

            if (data) response.end(data);else response.end();
            setTimeout(function () {
                callback();
            }, 1);
            return true;
        }

        /**
         * ends the streaming response
         *@param {string} filePath - the file path to serve.
         *@param {RServerResponse} - the response object
         *@param {number} status - the status code
         *@param {Object} headers - the request headers
         *@param {function} [callback] - a callback method
         *@returns {boolean}
        */

    }, {
        key: 'endStream',
        value: function endStream(filePath, response, status, headers, callback) {
            callback = Util.isCallable(callback) ? callback : function () {};

            response.writeHead(status, headers);

            var readStream = fs.createReadStream(filePath);

            readStream.on('end', function () {
                response.end(callback);
            }).on('error', function () {
                /* istanbul ignore next */
                readStream.end();
            });

            readStream.pipe(response, { end: false });
            return true;
        }

        /**
         * negotiates the content
        */

    }, {
        key: 'negotiateContent',
        value: function negotiateContent(headers, eTag, fileMTime) {
            if (typeof headers['if-none-match'] !== 'undefined' && headers['if-none-match'] === eTag) return true;

            if (typeof headers['if-modified-since'] !== 'undefined' && headers['if-modified-since'] === fileMTime) return true;

            return false;
        }

        /**
         * computes and returns a files eTag
         *@param {number} fileMTime - the files last modification time
         *@param {string} [length=16] - the hash tag length to generate
         *@returns {string}
        */

    }, {
        key: 'getFileTag',
        value: function getFileTag(fileMTime, length) {
            var hash = crypto.createHash('sha256');
            hash.update(fileMTime.toString());

            return hash.digest('hex').substring(0, length || 16);
        }

        /**
         * returns the directories default document if any
         *@param {string} dir - dir or file path.
         *@returns {string}
        */

    }, {
        key: 'getDefaultDocument',
        value: function getDefaultDocument(dir) {
            var _iteratorNormalCompletion = true;
            var _didIteratorError = false;
            var _iteratorError = undefined;

            try {
                for (var _iterator = this.defaultDocuments[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                    var file = _step.value;

                    if (fs.existsSync(path.join(dir, '/', file))) return file;
                }
            } catch (err) {
                _didIteratorError = true;
                _iteratorError = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion && _iterator.return) {
                        _iterator.return();
                    }
                } finally {
                    if (_didIteratorError) {
                        throw _iteratorError;
                    }
                }
            }

            return '';
        }

        /**
         * validates the request method and returns the public file or directory path that
         * matches the request url
         *@param {string} method - the request method
         *@returns {string}
        */

    }, {
        key: 'validateRequest',
        value: function validateRequest(url, method) {
            var validPath = '';
            //sanitize the url
            url = decodeURIComponent(url).replace(/[#?].*/, '').replace(/\.\./g, '');

            if (['GET', 'OPTIONS', 'HEAD'].includes(method.toUpperCase())) {
                var _iteratorNormalCompletion2 = true;
                var _didIteratorError2 = false;
                var _iteratorError2 = undefined;

                try {
                    for (var _iterator2 = this.publicPaths[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
                        var publicPath = _step2.value;

                        var testPath = path.join(publicPath, url);

                        if (fs.existsSync(testPath)) {

                            if (fs.statSync(testPath).isFile()) {
                                validPath = testPath;
                                break;
                            }

                            var defaultDocument = this.getDefaultDocument(testPath);
                            if (defaultDocument) {
                                validPath = path.join(testPath + '/' + defaultDocument);
                                break;
                            }
                        }
                    }
                } catch (err) {
                    _didIteratorError2 = true;
                    _iteratorError2 = err;
                } finally {
                    try {
                        if (!_iteratorNormalCompletion2 && _iterator2.return) {
                            _iterator2.return();
                        }
                    } finally {
                        if (_didIteratorError2) {
                            throw _iteratorError2;
                        }
                    }
                }
            }

            //do not serve files that starts with .
            if (validPath && path.basename(validPath).indexOf('.') !== 0) return validPath;else return '';
        }

        /**
         * serves a static file response back to the client
         *@param {string} url - the request url
         *@param {string} method - the request method
         *@param {Object} headers - the request headers
         *@param {RServerResponse} response - the response object
         *@param {Function} [callback] - a callback function that will be called once the operation
         * fails or completes
        */

    }, {
        key: 'serve',
        value: function serve(url, method, headers, response, callback) {
            method = method.toUpperCase();

            var filePath = this.validateRequest(url, method);
            if (filePath === '') return false;

            if (method === 'OPTIONS') return this.endResponse(response, 200, { 'Allow': 'OPTIONS, HEAD, GET, POST' }, null, callback);

            var resHeaders = this.getDefaultHeaders(filePath);

            if (this.negotiateContent(headers, resHeaders['ETag'], resHeaders['Last-Modified'])) return this.endResponse(response, 304, {}, null, callback);

            switch (method) {
                case 'HEAD':
                    resHeaders['Accept-Ranges'] = 'bytes';
                    return this.endResponse(response, 200, resHeaders, null, callback);

                case 'GET':
                    return this.endStream(filePath, response, 200, resHeaders, callback);
            }
        }

        /**
         * servers server http error files. such as 504, 404, etc
         *@param {RServerResponse} response - the response object
         *@param {number} status - the response status code
         *@param {string} baseDir - the user defined httErors base directory relative to root.
         *@param {string} filePath - the file path that is mapped to the error code
         *@param {Function} [callback] - a callback function that will be called once the operation
         * fails or completes
        */

    }, {
        key: 'serveHttpErrorFile',
        value: function serveHttpErrorFile(response, status, baseDir, filePath, callback) {
            if (!filePath) filePath = path.join(__dirname, '../httpErrors/' + status + '.html');else filePath = path.join(this.rootDir, baseDir, filePath);

            if (!fs.existsSync(filePath) || fs.statSync(filePath).isDirectory()) {
                response.statusCode = status;
                response.end(callback);
                return true;
            }

            var headers = this.getDefaultHeaders(filePath);
            this.endStream(filePath, response, status, headers, callback);
            return true;
        }

        /**
         * serves file intended for download to the client
         *@param {RServerResponse} response - the response object
         *@param {string} filePath - the file path
         *@param {string} [filename] - suggested file that the browser will use in saving the file
         *@param {Function} [callback] - a callback function that will be called once the operation
         * fails or completes
        */

    }, {
        key: 'serveDownload',
        value: function serveDownload(response, filePath, filename, callback) {
            var absPath = path.join(this.rootDir, filePath);
            if (!fs.existsSync(absPath) || fs.statSync(absPath).isDirectory()) {
                response.end(callback);
                return;
            }

            var resHeaders = this.getDefaultHeaders(filePath);
            filename = typeof filename === 'string' && filename ? filename : path.parse(absPath).base;

            resHeaders['Content-Disposition'] = 'attachment; filename="' + filename + '"';

            this.endStream(absPath, response, 200, resHeaders, callback);
        }
    }, {
        key: Symbol.toStringTag,
        get: function get$$1() {
            return 'StaticFileServer';
        }
    }]);
    return StaticFileServer;
}();

module.exports = StaticFileServer;
