'use strict';

var fs = require('fs');
var path = require('path');

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) {
  return typeof obj;
} : function (obj) {
  return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
};

var toConsumableArray = function (arr) {
  if (Array.isArray(arr)) {
    for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) arr2[i] = arr[i];

    return arr2;
  } else {
    return Array.from(arr);
  }
};

/**
 * Utility module
 * this module defines a bunch of utility functions that will be relevant to most other modules
*/
var toString = Object.prototype.toString;

var Util = {
    /**
     * tests if a variable is a number
     *@param {*} variable - variable to test
     *@returns {boolean}
    */
    isNumber: function isNumber(variable) {
        return typeof variable === 'number' && !isNaN(variable) && isFinite(variable);
    },


    /**
     * tests if a variable is a function
     *@param {*} variable - variable to test
     *@returns {boolean}
    */
    isCallable: function isCallable(variable) {
        return (toString.call(variable) === '[object Function]' || variable instanceof Function) && !(variable instanceof RegExp);
    },


    /**
     * tests if a variable is an array
     *@param {*} variable - variable to test
     *@returns {boolean}
    */
    isArray: function isArray(variable) {
        return toString.call(variable) === '[object Array]' || variable instanceof Array;
    },


    /**
     * tests if a variable is an object
     *@param {*} variable - variable to test
     *@returns {boolean}
    */
    isObject: function isObject(variable) {
        return (typeof variable === 'undefined' ? 'undefined' : _typeof(variable)) === 'object' && variable !== null;
    },


    /**
     * tests if a variable is a plain object literal
     *@param {*} variable - variable to test
     *@returns {boolean}
    */
    isPlainObject: function isPlainObject(variable) {
        if (this.isObject(variable)) {
            var prototypeOf = Object.getPrototypeOf(variable);
            return prototypeOf === null || prototypeOf === Object.getPrototypeOf({});
        }
        return false;
    },


    /**
     * tests if a variable is a valid function parameter
     *@param {*} variable - variable to test
     *@param {boolean} excludeNulls - boolean value indicating if null values should be
     * taken as an invalid parameter
     *@returns {boolean}
    */
    isValidParameter: function isValidParameter(variable, excludeNulls) {
        if (excludeNulls && variable === null) return false;
        return typeof variable !== 'undefined';
    },


    /**
     * returns the argument if it is already an array, or makes an array using the argument
     *@param {*} arg - the argument
     *@param {boolean} excludeNulls - boolean value indicating if null argument should default
     * to empty array just like undefined argument
     *@returns {Array}
    */
    makeArray: function makeArray(arg, excludeNulls) {
        if (this.isArray(arg)) return arg;
        return this.isValidParameter(arg, excludeNulls) ? [arg] : [];
    },

    /**
     * generates a callback function, scoping the execution with optional extra parameters
     *@param {Function} callback - the callback function
     *@param {Scope} [scope] - the execution scope - defaults to the host object
     *@param {Array} [parameters=null] - array of parameters to pass in to callback during execution
     *@throws {TypeError} throws error if callback is not a function
     *@returns {Function}
    */
    generateCallback: function generateCallback(callback, scope, parameters) {
        if (!this.isCallable(callback)) {
            throw new TypeError('argument one is not a function');
        }
        scope = this.isObject(scope) ? scope : null;
        parameters = this.makeArray(parameters);

        return function () {
            for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
                args[_key] = arguments[_key];
            }

            var mergedParameters = [].concat(args, toConsumableArray(parameters));
            try {
                return callback.apply(scope, mergedParameters);
            } catch (ex) {
                //
            }
        };
    },


    /**
     * runs the executable and supresses all runtime errors
     *@param {Function} executable - function to execute
     *@param {Scope} [scope] - runtime scope object - defaults to the host object
     *@param {Array} [parameters=null] - array of parameters to pass in to executable
     *@param {number} [runAfter=0] - least number of time in milliseconds to wait before
     * starting the execution
     *@throws {TypeError} if argument one is not a function
     *@returns {mixed} this will return a promise if runAfter parameter is given else it will
     * return the execution control
    */
    runSafe: function runSafe(executable, scope, parameters, runAfter) {
        var callback = this.generateCallback(executable, scope, parameters);
        if (runAfter) {
            return new Promise(function (resolve) {
                setTimeout(function () {
                    resolve(callback()); // pass in the return value to the resolve method
                }, runAfter);
            });
        }
        return callback();
    },


    /**
     * converts the letters into camel like cases
     *@param {string} value - the string word to convert
     *@param {string|RegExp} [delimiter=/[-_]/] - a delimiter string or regex pattern used in
     * finding split segments
     *@returns {string}
    */
    camelCase: function camelCase(value) {
        var delimiter = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : /[-_]/;

        value = value.toString();
        var tokens = value.split(delimiter).map(function (token, idx) {
            return idx === 0 ? token : token[0].toUpperCase() + token.substring(1);
        });
        return tokens.join('');
    },


    /**
     * generates a radom text with given length of characters
     *@param {number} [length=4] - char length of the random text to generate
     *@returns {string}
    */
    getRandomText: function getRandomText(length) {
        var chars = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789',
            rands = [],
            i = -1;

        length = length ? length : 4;
        while (++i < length) {
            rands.push(chars.charAt(Math.floor(Math.random() * chars.length)));
        }return rands.join('');
    },


    /**
     * performs a deep merge of all comma seperated list of objects and returns a new object
     *@param {...Object} objects - comma separated list of objects to merge
     *@returns {Object}
    */
    mergeObjects: function mergeObjects() {
        /**
         * runs the process
         *@param {Object} dest - the destination object
         *@param {Object} src - the src object
         *@returns {Object}
        */
        function run(dest, src) {
            var keys = Object.keys(src);
            var _iteratorNormalCompletion = true;
            var _didIteratorError = false;
            var _iteratorError = undefined;

            try {
                for (var _iterator = keys[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                    var key = _step.value;

                    var value = src[key];

                    if (typeof dest[key] === 'undefined') {
                        dest[key] = this.isPlainObject(value) ? run.call(this, {}, value) : value;
                    } else if (this.isPlainObject(value)) {
                        dest[key] = this.isPlainObject(dest[key]) ? run.call(this, dest[key], value) : run.call(this, {}, value);
                    } else {
                        dest[key] = value;
                    }
                }
            } catch (err) {
                _didIteratorError = true;
                _iteratorError = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion && _iterator.return) {
                        _iterator.return();
                    }
                } finally {
                    if (_didIteratorError) {
                        throw _iteratorError;
                    }
                }
            }

            return dest;
        }

        var dest = {};

        for (var _len2 = arguments.length, objects = Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
            objects[_key2] = arguments[_key2];
        }

        var _iteratorNormalCompletion2 = true;
        var _didIteratorError2 = false;
        var _iteratorError2 = undefined;

        try {
            for (var _iterator2 = objects[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
                var object = _step2.value;

                if (!this.isPlainObject(object)) continue;
                dest = run.call(this, dest, object);
            }
        } catch (err) {
            _didIteratorError2 = true;
            _iteratorError2 = err;
        } finally {
            try {
                if (!_iteratorNormalCompletion2 && _iterator2.return) {
                    _iterator2.return();
                }
            } finally {
                if (_didIteratorError2) {
                    throw _iteratorError2;
                }
            }
        }

        return dest;
    },


    /**
     * creates a directory recursively and synchronously
     *@param {string} dir - the directory to create
     *@returns {boolean}
    */
    mkDirSync: function mkDirSync(dir) {
        if (typeof dir !== 'string') throw new TypeError('argument one is not a string');
        if (dir === '/' || dir === '' || fs.existsSync(dir)) return false;

        //search backwards
        dir = dir.replace(/\/+$/, '');
        var existingPath = '',
            testPath = dir;
        while (existingPath === '' && testPath !== '/') {
            testPath = path.join(testPath, '../');
            if (fs.existsSync(testPath)) existingPath = testPath;
        }

        var pathTokens = dir.split(existingPath)[1].split('/');
        var _iteratorNormalCompletion3 = true;
        var _didIteratorError3 = false;
        var _iteratorError3 = undefined;

        try {
            for (var _iterator3 = pathTokens[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
                var pathToken = _step3.value;

                existingPath = path.join(existingPath, '/', pathToken);
                fs.mkdirSync(existingPath);
            }
        } catch (err) {
            _didIteratorError3 = true;
            _iteratorError3 = err;
        } finally {
            try {
                if (!_iteratorNormalCompletion3 && _iterator3.return) {
                    _iterator3.return();
                }
            } finally {
                if (_didIteratorError3) {
                    throw _iteratorError3;
                }
            }
        }

        return true;
    }
};

module.exports = Util;
